/* $Id: copywindow.c,v 1.2 2006/01/28 16:46:10 cegger Exp $
***************************************************************************

   X server for LibGGI - Graphics acceleration

   Copyright (C) 1997	   Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998-1999 Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/

#include "screen.h"


/*
 * Ripped from hw/xfree86/accel/s3/s3blt.c
 */

static void
xggiFindOrdering(DrawablePtr pSrcDrawable, DrawablePtr pDstDrawable, GC *pGC,
		 int numRects, BoxPtr boxes, int srcx, int srcy,
		 int dstx, int dsty, unsigned int *ordering)
{
	int   i, j, y;
	int   xMax, yMin, yMax;

	/*
	 * If not the same drawable then order of move doesn't matter. Following
	 * assumes that boxes are sorted from top to bottom and left to right.
	 */
	if ((pSrcDrawable != pDstDrawable) &&
	    ((pGC->subWindowMode != IncludeInferiors) ||
	     (pSrcDrawable->type == DRAWABLE_PIXMAP) ||
	     (pDstDrawable->type == DRAWABLE_PIXMAP))) {
		for (i = 0; i < numRects; i++)
			ordering[i] = i;
	} else {	/* within same drawable, must sequence moves
			 * carefully! */
		if (dsty <= srcy) {	/* Scroll up or stationary vertical. Vertical
					 * order OK */
			if (dstx <= srcx) {	/* Scroll left or stationary horizontal.
						 * Horizontal order OK as well */
				for (i = 0; i < numRects; i++) {
					ordering[i] = i;
				}	/* for */
			} else {	/* scroll right. must reverse horizontal
					 * banding of rects. */
				for (i = 0, j = 1, xMax = 0; i < numRects; j = i + 1, xMax = i) {
					/* find extent of current horizontal band */
					y = boxes[i].y1;	/* band has this y coordinate */
					while ((j < numRects) && (boxes[j].y1 == y))
						j++;
					/* reverse the horizontal band in the output ordering */
					for (j--; j >= xMax; j--, i++)
						ordering[i] = j;
				}	/* for */
			}	/* if */
		} else {			/* Scroll down. Must reverse vertical
						 * banding. */
			if (dstx < srcx) {	/* Scroll left. Horizontal order OK. */
				for (i = numRects - 1, j = i - 1, yMin = i,
					     yMax = 0; i >= 0;
				     j = i - 1, yMin = i) {
					/* find extent of current horizontal
					   band */
					/* band has this y coordinate */
					y = boxes[i].y1;
					while ((j >= 0) && (boxes[j].y1 == y))
						j--;
					/* reverse the horizontal band in the
					   output ordering */
					for (j++; j <= yMin; j++, i--, yMax++) {
						ordering[yMax] = j;
					}	/* for */
				}	/* for */
			} else {
				/* Scroll right or horizontal stationary.
				 * Reverse horizontal order as well (if
				 * stationary, horizontal order can be
				 * swapped without penalty and this is faster
				 * to compute).
				 */
				for (i = 0, j = numRects - 1; i < numRects;
				     i++, j--)
				{
					ordering[i] = j;
				}	/* for */
			}	/* if */
		}	/* if */
	}	/* if */
}	/* xggiFindOrdering */


/*
 * xggiCopyWindow - accelerated replacement for cfbCopyWindow
 *
 * pWin->drawable.x/y contains the destination position
 * ptOldOrg contains the old position
 * prgnSrc is a region describing the part of the source to be moved
 *  (this is padded by one pixel at all four edges of a window?!?!?)
 *
 * I don't really understand this, BTW ;)
 *
 */
void
xggiCopyWindow(WindowPtr pWin, DDXPointRec ptOldOrg, RegionPtr prgnSrc)
{
	RegionRec rgnDst;
	BoxPtr pboxOrig, pbox;
	int dx, dy;
	int i, nbox;
	unsigned int *ordering;
	GC    dummyGC;

	dummyGC.subWindowMode = ~(unsigned)IncludeInferiors;
    
	REGION_NULL(pWin->drawable.pScreen, &rgnDst);

	dx = ptOldOrg.x - pWin->drawable.x;
	dy = ptOldOrg.y - pWin->drawable.y;

	REGION_TRANSLATE(pWin->drawable.pScreen, prgnSrc, -dx, -dy);
	REGION_INTERSECT(pWin->drawable.pScreen, &rgnDst,
			 &pWin->borderClip, prgnSrc);

	pboxOrig = REGION_RECTS(&rgnDst);
	nbox = REGION_NUM_RECTS(&rgnDst);

	ordering = (unsigned int *)ALLOCATE_LOCAL(nbox * sizeof(unsigned int));

	if (ordering) {
		int scr = xggiGetScreenIdx(pWin->drawable.pScreen);

		if (!(WINDOW_HAS_MASTER_VISUAL_BPP(pWin, scr))) {
			ErrorF("BUG in xggiCopyWindow(): drawable depth %i != visual depth %i\n",
				pWin->drawable.bitsPerPixel,
				MASTER_VISUAL_BPP(scr));
		}

		xggiFindOrdering((DrawablePtr)pWin, (DrawablePtr)pWin,
				 &dummyGC, nbox, pboxOrig,
				 ptOldOrg.x, ptOldOrg.y,
				 pWin->drawable.x, pWin->drawable.y, ordering);
		DBUF_PROLOGUE(scr);
		for (i = 0; i < nbox; i++) {
			pbox = &pboxOrig[ordering[i]];
			ggiCopyBox(MASTER_VISUAL_VIS(scr),
				   pbox->x1 + dx, pbox->y1 + dy,
				   pbox->x2 - pbox->x1, pbox->y2 - pbox->y1,
				   pbox->x1, pbox->y1);
		}	/* for */
		DBUF_EPILOGUE(scr);
		DEALLOCATE_LOCAL(ordering);
	}	/* if */

	REGION_UNINIT(pWin->drawable.pScreen, &rgnDst);
}	/* xggiCopyWindow */
